/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/def.hpp>
#include <ox/std/error.hpp>
#include <ox/std/string.hpp>

#include <ox/claw/write.hpp>

#include "gfx.hpp"

namespace nostalgia::core {

class ModelContainer {
	public:
		virtual ~ModelContainer() = default;
		virtual void *obj() noexcept = 0;
		virtual const char *typeName() noexcept = 0;
		virtual int typeVersion() noexcept = 0;
};

template<typename T>
class ModelContainerTemplate: public ModelContainer {
	private:
		T m_obj;

	public:
		constexpr ModelContainerTemplate() = default;

		void *obj() noexcept final {
			return &m_obj;
		}

		const char *typeName() noexcept final {
			return T::TypeName;
		}

		int typeVersion() noexcept final {
			return T::TypeVersion;
		}

};

using ClawObjPtr = ox::UniquePtr<ModelContainer>;

struct BaseConverter {
	virtual ~BaseConverter() noexcept = default;

	[[nodiscard]]
	virtual ox::String srcTypeName() noexcept = 0;

	[[nodiscard]]
	virtual int srcTypeVersion() noexcept = 0;

	[[nodiscard]]
	virtual bool dstMatches(const ox::String &dstTypeName, int dstTypeVersion) noexcept = 0;

	[[nodiscard]]
	virtual bool matches(const ox::String &srcTypeName, int srcTypeVersion, const ox::String &dstTypeName, int dstTypeVersion) noexcept = 0;

	virtual ox::Result<ClawObjPtr> convertPtrToPtr(const ClawObjPtr &src) noexcept = 0;

	virtual ox::Result<ClawObjPtr> convertBuffToPtr(const ox::Buffer &srcBuff) noexcept = 0;
};

template<typename SrcType, typename DstType>
struct Converter: public BaseConverter{
	ox::String srcTypeName() noexcept final {
		return SrcType::TypeName;
	}

	int srcTypeVersion() noexcept final {
		return SrcType::TypeVersion;
	}

	bool dstMatches(const ox::String &dstTypeName, int dstTypeVersion) noexcept final {
		return dstTypeName == DstType::TypeName
		    && dstTypeVersion == DstType::TypeVersion;
	}

	bool matches(const ox::String &srcTypeName, int srcTypeVersion, const ox::String &dstTypeName, int dstTypeVersion) noexcept final {
		return srcTypeName == SrcType::TypeName
		    && dstTypeName == DstType::TypeName
		    && srcTypeVersion == SrcType::TypeVersion
		    && dstTypeVersion == DstType::TypeVersion;
	}

	virtual ox::Error convert(SrcType*, DstType*) noexcept {
		return OxError(1);
	}

	ox::Result<ClawObjPtr> convertPtrToPtr(const ClawObjPtr &src) noexcept final {
		auto dst = ox::make_unique<ModelContainerTemplate<DstType>>();
		oxReturnError(convert(reinterpret_cast<SrcType *>(src->obj()), reinterpret_cast<DstType *>(dst->obj())));
		return ox::Result<ClawObjPtr>(std::move(dst));
	}

	ox::Result<ClawObjPtr> convertBuffToPtr(const ox::Buffer &srcBuff) noexcept final {
		oxRequireM(src, ox::readClaw<SrcType>(srcBuff));
		auto dst = ox::make_unique<ModelContainerTemplate<DstType>>();
		oxReturnError(convert(&src, reinterpret_cast<DstType*>(dst->obj())));
		return ox::Result<ClawObjPtr>(std::move(dst));
	}
};

ox::Result<ClawObjPtr> convert(const ox::Buffer &srcBuffer, const ox::String &dstTypeName, int dstTypeVersion) noexcept;

template<typename DstType>
ox::Result<ClawObjPtr> convert(const ox::Buffer &srcBuffer) noexcept {
	return convert(srcBuffer, DstType::TypeName, DstType::TypeVersion);
}

template<typename T>
ox::Error convert(const ox::Buffer &buff, T *outObj) noexcept {
	oxRequireM(outPtr, convert(buff, T::TypeName, T::TypeVersion));
	*outObj = std::move(*reinterpret_cast<T*>(outPtr->obj()));
	return OxError(0);
}

template<typename DstType>
ox::Result<ox::Buffer> convertBuffToBuff(const ox::Buffer &srcBuffer, ox::ClawFormat fmt) noexcept {
	oxRequireM(out, convert(srcBuffer, DstType::TypeName, DstType::TypeVersion));
	return ox::writeClaw<DstType>(reinterpret_cast<DstType*>(out->obj()), fmt);
}

}
